@@ -1,13 +1,14 @@
 package Stash::REST;
 use strict;
 use 5.008_005;
-our $VERSION = '0.04';
+our $VERSION = '0.05';
 
 use warnings;
 use utf8;
 use URI;
+use URI::QueryParam;
 use HTTP::Request::Common qw(GET POST DELETE HEAD);
-use Carp qw/confess/;
+use Carp qw/confess cluck/;
 
 use Moo;
 use namespace::clean;
@@ -15,26 +16,26 @@ use namespace::clean;
 use Class::Trigger;
 
 has 'do_request' => (
-    is => 'rw',
-    isa => sub {die "$_[0] is not a CodeRef" unless ref $_[0] eq 'CODE'},
+    is       => 'rw',
+    isa      => sub { die "$_[0] is not a CodeRef" unless ref $_[0] eq 'CODE' },
     required => 1
 );
 
 has 'decode_response' => (
-    is => 'rw',
-    isa => sub {die "$_[0] is not a CodeRef" unless ref $_[0] eq 'CODE'},
+    is       => 'rw',
+    isa      => sub { die "$_[0] is not a CodeRef" unless ref $_[0] eq 'CODE' },
     required => 1
 );
 
 has 'stash' => (
-    is => 'rw',
-    isa => sub {die "$_[0] is not a HashRef" unless ref $_[0] eq 'HASH'},
+    is  => 'rw',
+    isa => sub { die "$_[0] is not a HashRef" unless ref $_[0] eq 'HASH' },
     default => sub { {} }
 );
 
 has 'fixed_headers' => (
-    is => 'rw',
-    isa => sub {die "$_[0] is not a ArrayRef" unless ref $_[0] eq 'ARRAY'},
+    is  => 'rw',
+    isa => sub { die "$_[0] is not a ArrayRef" unless ref $_[0] eq 'ARRAY' },
     default => sub { [] }
 );
 
@@ -57,96 +58,110 @@ around 'stash' => sub {
 };
 
 sub _capture_args {
-    my ($self, @params) = @_;
-    my ($uri, $data, %conf);
+    my ( $method, $self, @params ) = @_;
+    my ( $uri, $data, %conf );
 
-    confess 'rest_post invalid number of params' if @params < 1;
+    confess 'invalid number of params' if @params < 1;
 
     $uri = shift @params;
-    confess 'rest_post invalid uri param' if ref $uri ne '' && ref $uri ne 'ARRAY';
+    confess 'invalid uri param' if ref $uri ne '' && ref $uri ne 'ARRAY';
 
     $uri = join '/', @$uri if ref $uri eq 'ARRAY';
 
-    if (scalar @params % 2 == 0){
+    # if number of params is odd, then, the last item is defined as $data
+    if ( scalar @params % 2 == 0 ) {
         %conf = @params;
-        $data = exists $conf{data} ? $conf{data} : [];
-    }else{
+        $data = exists $conf{data} ? $conf{data} : undef;
+    }
+    else {
         $data = pop @params;
         %conf = @params;
     }
 
-    confess 'rest_post param $data invalid' unless ref $data eq 'ARRAY';
+    confess 'param $data should be an array ref'
+      if ref $data ne 'ARRAY' && ( exists $conf{headers} && !grep { 'Content-Type' } @{ $conf{headers} } );
 
+    confess "Can't use ->{files} helper with custom Content-Type."
+      if exists $conf{files} && ( exists $conf{headers} && grep { 'Content-Type' } @{ $conf{headers} } );
 
-    return ($self, $uri, $data, %conf);
+    my $can_have_body = $method =~ /POST|PUT/;
+
+    if ( !$can_have_body && $data && ref $data eq 'ARRAY' && @$data ) {
+        confess "$method can't have {data}. Please, use only {params} instead" if ( exists $conf{params} );
+
+        $conf{params} = $data;
+
+    }
+    elsif ( !$can_have_body && $data ) {
+        cluck "$method does not allow body. You may have problems with proxy. Consider removing it";
+        $conf{data} = $data;
+    }
+    else {
+        $conf{data} = $data;
+    }
+
+    return ( $self, $uri, %conf );
 }
 
 sub rest_put {
-    my ($self, $url, $data, %conf) = &_capture_args(@_);
+    my ( $self, $url, %conf ) = &_capture_args( 'PUT', @_ );
 
-    $self->call_trigger('before_rest_put', \$url, \$data, \%conf);
+    $self->call_trigger( 'before_rest_put', { url => $url, conf => \%conf } );
     $self->_rest_request(
         $url,
         code => ( exists $conf{is_fail} ? 400 : 202 ),
         %conf,
-        method => 'PUT',
-        $data
+        method => 'PUT'
     );
 }
 
 sub rest_head {
-    my ($self, $url, $data, %conf) = &_capture_args(@_);
+    my ( $self, $url, %conf ) = &_capture_args( 'HEAD', @_ );
 
-    $self->call_trigger('before_rest_head', \$url, \$data, \%conf);
+    $self->call_trigger( 'before_rest_head', { url => $url, conf => \%conf } );
     $self->_rest_request(
         $url,
         code => 200,
         %conf,
-        method => 'HEAD',
-        $data
+        method => 'HEAD'
     );
 }
 
 sub rest_delete {
-    my ($self, $url, $data, %conf) = &_capture_args(@_);
+    my ( $self, $url, %conf ) = &_capture_args( 'DELETE', @_ );
 
-    $self->call_trigger('before_rest_delete', \$url, \$data, \%conf);
+    $self->call_trigger( 'before_rest_delete', { url => $url, conf => \%conf } );
     $self->_rest_request(
         $url,
         code => 204,
         %conf,
-        method => 'DELETE',
-        $data
+        method => 'DELETE'
     );
 }
 
 sub rest_get {
-    my ($self, $url, $data, %conf) = &_capture_args(@_);
+    my ( $self, $url, %conf ) = &_capture_args( 'GET', @_ );
 
-    $self->call_trigger('before_rest_get', \$url, \$data, \%conf);
+    $self->call_trigger( 'before_rest_get', { url => $url, conf => \%conf } );
     $self->_rest_request(
         $url,
         code => 200,
         %conf,
-        method => 'GET',
-        $data
+        method => 'GET'
     );
 }
 
 sub rest_post {
-    my ($self, $url, $data, %conf) = &_capture_args(@_);
-    $self->call_trigger('before_rest_post', \$url, \$data, \%conf);
+    my ( $self, $url, %conf ) = &_capture_args( 'POST', @_ );
+    $self->call_trigger( 'before_rest_post', { url => $url, conf => \%conf } );
 
-    $self->_rest_request(
-        $url,
-        %conf,
-        method => 'POST',
-        $data
-    );
+    $self->_rest_request( $url, %conf, method => 'POST' );
 }
 
 sub _rest_request {
-    my ($self, $url, $data, %conf) = &_capture_args(@_);
+    my ( $self, $url, %conf ) = @_;
+
+    my $data = exists $conf{data} ? $conf{data} : undef;
 
     my $is_fail = exists $conf{is_fail} && $conf{is_fail};
 
@@ -154,48 +169,57 @@ sub _rest_request {
     $code ||= $is_fail ? 400 : 201;
     $conf{code} = $code;
 
+    my $uri = URI->new($url);
+    $uri->query_param_append( @{ $conf{params} } ) if $conf{params};
+    $url = $uri->as_string;
+
     my $stashkey = exists $conf{stash} ? $conf{stash} : undef;
 
-    my @headers = (@{$self->fixed_headers()}, @{$conf{headers}||[]} );
+    my @headers = ( @{ $self->fixed_headers() }, @{ $conf{headers} || [] } );
 
     my $req;
 
     if ( !exists $conf{files} ) {
-        $req = POST $url, $data, @headers;
+        if ( defined $data ) {
+            $req = POST $url, @headers, Content => $data;
+        }
+        else {
+            $req = GET $url, @headers;
+        }
     }
     else {
         $conf{files}{$_} = [ $conf{files}{$_} ] for keys %{ $conf{files} };
 
-        $req = POST $url,
-          @headers,
+        $req = POST $url, @headers,
           'Content-Type' => 'form-data',
-          Content => [ @$data, %{ $conf{files} } ];
+          Content        => [ ( $data && ref $data eq 'ARRAY' ? @$data : () ), %{ $conf{files} } ];
     }
 
-    $self->call_trigger('process_request', \$req, \%conf);
+    $self->call_trigger( 'process_request', { req => $req, conf => \%conf } );
 
-    $req->method( $conf{method} ) if exists $conf{method};
+    # change to correct method.
+    $req->method( $conf{method} );
 
-    my $res = eval{$self->do_request()->($req)};
+    my $res = eval { $self->do_request()->($req) };
     confess "request died: $@" if $@;
 
-    $self->call_trigger('process_response', \$req, \$res, \%conf);
+    $self->call_trigger( 'process_response', { req => $req, res => $res, conf => \%conf } );
 
     #is( $res->code, $code, $name . ' status code is ' . $code );
-    confess 'response expected fail and it is successed' if $is_fail && $res->is_success;
+    confess 'response expected fail and it is successed' if $is_fail  && $res->is_success;
     confess 'response expected success and it is failed' if !$is_fail && !$res->is_success;
 
-    confess 'response code [',$res->code,'] diverge expected [',$code,']' if $code != $res->code;
+    confess 'response code [', $res->code, '] diverge expected [', $code, ']' if $code != $res->code;
 
-    $self->call_trigger('process_response_success', \$req, \$res, \%conf);
+    $self->call_trigger( 'process_response_success', { req => $req, res => $res, conf => \%conf } );
 
     return '' if $code == 204;
     return $res if exists $conf{method} && $conf{method} eq 'HEAD';
 
-    my $obj = eval { $self->decode_response()->( $res ) };
+    my $obj = eval { $self->decode_response()->($res) };
     confess("decode_response failed: $@") if $@;
 
-    $self->call_trigger('response_decoded', \$req, \$res, \$obj, \%conf);
+    $self->call_trigger( 'response_decoded', { req => $req, res => $res, decoded => $obj, conf => \%conf } );
 
     if ($stashkey) {
         $self->stash->{$stashkey} = $obj;
@@ -207,17 +231,18 @@ sub _rest_request {
 
             my $item_url = $res->header('Location');
 
-            if ($item_url){
-                $self->stash->{$stashkey . '.url'} = $item_url ;
+            if ($item_url) {
+                $self->stash->{ $stashkey . '.url' } = $item_url;
 
                 $self->rest_reload($stashkey);
 
-                $self->call_trigger('item_loaded', $stashkey, \%conf);
-            }else{
+                $self->call_trigger( 'item_loaded', { stash => $stashkey, conf => \%conf } );
+            }
+            else {
                 confess 'requests with response code 201 should contain header Location';
             }
 
-            $self->call_trigger('stash_added', $stashkey, \%conf);
+            $self->call_trigger( 'stash_added', { stash => $stashkey, conf => \%conf } );
         }
     }
 
@@ -227,14 +252,13 @@ sub _rest_request {
 
         $self->rest_reload_list($stashkey);
 
-        $self->call_trigger('list_loaded', $stashkey, \%conf);
+        $self->call_trigger( 'list_loaded', { stash => $stashkey, conf => \%conf } );
 
     }
 
     return $obj;
 }
 
-
 sub rest_reload {
     my $self     = shift;
     my $stashkey = shift;
@@ -244,8 +268,7 @@ sub rest_reload {
     my $code = exists $conf{code} ? $conf{code} : 200;
     $conf{code} = $code;
 
-
-    my @headers = (@{$self->fixed_headers()}, @{$conf{headers}||[]} );
+    my @headers = ( @{ $self->fixed_headers() }, @{ $conf{headers} || [] } );
     my $item_url = $self->stash->{ $stashkey . '.url' };
 
     confess "can't stash $stashkey.url is not valid" unless $item_url;
@@ -256,38 +279,38 @@ sub rest_reload {
       : undef;
 
     confess 'prepare_request must be a coderef'
-        if $prepare_request && ref $prepare_request ne 'CODE';
+      if $prepare_request && ref $prepare_request ne 'CODE';
 
     my $req = POST $item_url, @headers, [];
     $req->method('GET');
     $prepare_request->($req) if $prepare_request;
 
-    $self->call_trigger('process_request', \$req, \%conf);
+    $self->call_trigger( 'process_request', { req => $req, conf => \%conf } );
     my $res = $self->do_request()->($req);
 
-    $self->call_trigger('process_response', \$req, \$res, \%conf);
+    $self->call_trigger( 'process_response', { req => $req, res => $res, conf => \%conf } );
 
     confess 'request code diverge expected' if $code != $res->code;
 
-    $self->call_trigger('process_response_success', \$req, \$res, \%conf);
+    $self->call_trigger( 'process_response_success', { req => $req, res => $res, conf => \%conf } );
 
     my $obj;
     if ( $res->code == 200 ) {
-        my $obj = eval { $self->decode_response()->( $res ) };
+        my $obj = eval { $self->decode_response()->($res) };
         confess("decode_response failed: $@") if $@;
 
-        $self->call_trigger('response_decoded', \$req, \$res, \$obj, \%conf);
+        $self->call_trigger( 'response_decoded', { req => $req, res => $res, decoded => $obj, conf => \%conf } );
 
         $self->stash( $stashkey . '.get' => $obj );
     }
     elsif ( $res->code == 404 ) {
 
-        $self->call_trigger('stash_removed', $stashkey, \%conf);
+        $self->call_trigger( 'stash_removed', { stash => $stashkey, conf => \%conf } );
 
         # $self->stash->{ $stashkey . '.get' };
         delete $self->stash->{ $stashkey . '.id' };
         delete $self->stash->{ $stashkey . '.url' };
-        delete $self->stash->{ $stashkey };
+        delete $self->stash->{$stashkey};
 
     }
     else {
@@ -297,7 +320,6 @@ sub rest_reload {
     return $obj;
 }
 
-
 sub rest_reload_list {
     my $self     = shift;
     my $stashkey = shift;
@@ -307,7 +329,7 @@ sub rest_reload_list {
     my $code = exists $conf{code} ? $conf{code} : 200;
     $conf{code} = $code;
 
-    my @headers = (@{$self->fixed_headers()}, @{$conf{headers}||[]} );
+    my @headers = ( @{ $self->fixed_headers() }, @{ $conf{headers} || [] } );
     my $item_url = $self->stash->{ $stashkey . '.list-url' };
 
     confess "can't stash $stashkey.list-url is not valid" unless $item_url;
@@ -317,28 +339,28 @@ sub rest_reload_list {
       ? $self->stash->{ $stashkey . '.prepare_request' }
       : undef;
     confess 'prepare_request must be a coderef'
-        if $prepare_request && ref $prepare_request ne 'CODE';
+      if $prepare_request && ref $prepare_request ne 'CODE';
 
     my $req = POST $item_url, @headers, [];
     $req->method('GET');
     $prepare_request->($req) if $prepare_request;
 
-    $self->call_trigger('process_request', \$req, \%conf);
+    $self->call_trigger( 'process_request', { req => $req, conf => \%conf } );
 
     my $res = $self->do_request()->($req);
 
-    $self->call_trigger('process_response', \$req, \$res, \%conf);
+    $self->call_trigger( 'process_response', { req => $req, res => $res, conf => \%conf } );
 
     confess 'request code diverge expected' if $code != $res->code;
 
-    $self->call_trigger('process_response_success', \$req, \$res, \%conf);
+    $self->call_trigger( 'process_response_success', { req => $req, res => $res, conf => \%conf } );
 
     my $obj;
     if ( $res->code == 200 ) {
-        my $obj = eval { $self->decode_response()->( $res ) };
+        my $obj = eval { $self->decode_response()->($res) };
         confess("decode_response failed: $@") if $@;
 
-        $self->call_trigger('response_decoded', \$req, \$res, \$obj, \%conf);
+        $self->call_trigger( 'response_decoded', { req => $req, res => $res, decoded => $obj, conf => \%conf } );
 
         $self->stash( $stashkey . '.list' => $obj );
     }
@@ -352,15 +374,14 @@ sub rest_reload_list {
 sub stash_ctx {
     my ( $self, $staname, $sub ) = @_;
 
-    $self->call_trigger('before_stash_ctx', $staname);
+    $self->call_trigger( 'before_stash_ctx', { stash => $staname } );
 
     my @ret = $sub->( $self->stash->{$staname} );
 
-    $self->call_trigger('after_stash_ctx', $staname, \@ret);
+    $self->call_trigger( 'after_stash_ctx', { stash => $staname, results => \@ret } );
     return @ret;
 }
 
-
 1;
 
 __END__
@@ -506,66 +527,63 @@ extensions and analysis by other modules, using the callbacks L<Class::Trigger>.
 
 =head2 rest_get
 
-Same as:
+Similar to:
 
     $self->rest_post(
         $url,
         code => 200,
         %conf,
-        method => 'GET',
-        $data
+        method => 'GET'
     );
 
 =head2 rest_put
 
-Same as:
+Similar to:
 
     $self->rest_post(
         $url,
         code => ( exists $conf{is_fail} ? 400 : 202 ),
         %conf,
         method => 'PUT',
-        $data
     );
 
 =head2 rest_head
 
-Same as:
+Similar to:
 
     $self->rest_post(
         $url,
         code => 200,
         %conf,
         method => 'HEAD',
-        $data
     );
 
 =head2 rest_delete
 
-Same as:
+Similar to:
 
     $self->rest_post(
         $url,
         code => 204,
         %conf,
-        method => 'DELETE',
-        $data
+        method => 'DELETE'
     );
 
 =head2 _capture_args
 
 This is a private method. It parse and validate params for rest_post and above methods.
 
-    my ($self, $url, $data, %conf) = &_capture_args(@_);
+    my ($self, $url, %conf) = &_capture_args(@_);
 
     So, @_ must have a items like this:
 
         $self, # (required) self object
         $url,  # (required) a string. If ARRAY $url will return a join '/', @$url
         %conf, # (optional) configuration hash (AKA list). Odd number will cause problems.
-        $data  # (optional) ArrayRef, send on body as application/x-www-form-urlencoded data
+        $data  # (optional) ArrayRef, send on body as application/x-www-form-urlencoded data on POST | PUT
+
+    # $data can be also sent as $conf{data} = [ ... ]
 
-    # $data can be also sent as $conf{data} = []
 
 =head2 rest_post
 
@@ -592,6 +610,50 @@ Load parsed response on C< $obj->stash->{foobar} > and some others fields
 C< $obj->stash->{foobar.id} > if response code is 201 and parsed response contains ->{id}
 C< $obj->stash->{foobar.url} > if response code is 201 and header contains Location (confess if missed)
 
+=head4 params => []
+
+Add query params to URI
+
+    $obj->rest_post(
+        '/foo?you_can_put_here=1',
+        list  => 1,
+        stash => 'easyname',
+        params => [ and_continue_with_other_params => 'here' ],
+        data => [ data_here_are_sent_on_body => '1' ]
+    );
+
+=head4 headers => [],
+
+Add headers on the main request.
+
+You can use this to modify the default HTTP::Request::Common header for $data
+
+    # POST body with JSON:
+    $obj->rest_post(
+        '/post-with-params?api_key=1',
+        stash   => 'testparams',
+        code => 200,
+        params  => [ another_key => 2 ],
+        headers => [
+            'Content-Type' => 'application/json',
+        ],
+        data => encode_json( { hello => 'json' } ),
+    );
+
+=head4 prepare_request => sub {}
+
+Modify secondary requests like GET /foo or GET Location after a POST /foo
+
+    $obj->rest_post(
+        '/foo',
+        list  => 1,
+        stash => 'easyname',
+        prepare_request => sub {
+            is(ref $_[0], 'HTTP::Request', 'HTTP::Request recived on prepare_request');
+            $run++;
+        },
+        data => [ name => 'foo' ]
+    );
 
 =head4 list => 1,
 
@@ -646,17 +708,42 @@ passing arguments. Unlike catalyst, it's never cleared, so, it lasts until objec
 I'm always trying to improve those numbers.
 Improve branch number is a very time-consuming task. There is a room for test all checking and defaults on tests.
 
-
+    @ version 0.05
     ---------------------------- ------ ------ ------ ------ ------ ------ ------
     File                           stmt   bran   cond    sub    pod   time  total
     ---------------------------- ------ ------ ------ ------ ------ ------ ------
-    blib/lib/Stash/REST.pm         94.9   73.1   81.4  100.0    0.0  100.0   84.6
-    Total                          94.9   73.1   81.4  100.0    0.0  100.0   84.6
+    blib/lib/Stash/REST.pm         96.4   74.5   68.4  100.0  100.0  100.0   86.7
+    Total                          96.4   74.5   68.4  100.0  100.0  100.0   86.7
     ---------------------------- ------ ------ ------ ------ ------ ------ ------
 
+
 =head1 Class::Trigger names
 
-Updated @ Stash-REST 0.04
+
+Updated @ Stash-REST 0.05
+
+    $ grep  '$self0_05->call_trigger' lib/Stash/REST.pm  | perl -ne '$_ =~ s/^\s+//; $_ =~ s/self-/self0_04-/; print' | sort | uniq
+
+Trigger / variables:
+
+    $self0_05->call_trigger( 'after_stash_ctx', { stash => $staname, results => \@ret });
+    $self0_05->call_trigger( 'before_rest_delete', { url => $url, conf => \%conf } );
+    $self0_05->call_trigger( 'before_rest_get', { url => $url, conf => \%conf } );
+    $self0_05->call_trigger( 'before_rest_head', { url => $url, conf => \%conf } );
+    $self0_05->call_trigger( 'before_rest_post', { url => $url, conf => \%conf } );
+    $self0_05->call_trigger( 'before_rest_put', { url => $url, conf => \%conf } );
+    $self0_05->call_trigger( 'before_stash_ctx', { stash => $staname } );
+    $self0_05->call_trigger( 'item_loaded', { stash => $stashkey, conf => \%conf });
+    $self0_05->call_trigger( 'list_loaded', { stash => $stashkey, conf => \%conf } );
+    $self0_05->call_trigger( 'process_request', {req => $req, conf => \%conf} );
+    $self0_05->call_trigger( 'process_response', {req => $req, res => $res, conf => \%conf} );
+    $self0_05->call_trigger( 'process_response_success', {req => $req, res => $res, conf => \%conf} );
+    $self0_05->call_trigger( 'response_decoded', {req => $req, res => $res, decoded => $obj, conf => \%conf} );
+    $self0_05->call_trigger( 'stash_added', { stash => $stashkey, conf => \%conf } );
+    $self0_05->call_trigger( 'stash_removed', { stash => $stashkey, conf => \%conf } );
+
+
+Updated @ Stash-REST 0.04 *** PLEASE UPGRADE TO NEW VERSION ***
 
     $ grep  '$self->call_trigger' lib/Stash/REST.pm  | perl -ne '$_ =~ s/^\s+//; $_ =~ s/self-/self0_04-/; print' | sort | uniq
 
@@ -679,26 +766,7 @@ Updated @ Stash-REST 0.04
     $self0_04->call_trigger('stash_removed', $stashkey, \%conf);
 
 
-Updated @ Stash-REST 0.03
 
-    $ grep  '$self->call_trigger' lib/Stash/REST.pm  | perl -ne '$_ =~ s/^\s+//; $_ =~ s/self-/self0_03-/; print' | sort | uniq
-
-    Trigger / variables:
-
-    $self0_03->call_trigger('after_stash_ctx', $staname, \@ret);
-    $self0_03->call_trigger('before_rest_delete', \$url, \$data, \%conf);
-    $self0_03->call_trigger('before_rest_get', \$url, \$data, \%conf);
-    $self0_03->call_trigger('before_rest_head', \$url, \$data, \%conf);
-    $self0_03->call_trigger('before_rest_post', \$url, \$data, \%conf);
-    $self0_03->call_trigger('before_rest_put', \$url, \$data, \%conf);
-    $self0_03->call_trigger('before_stash_ctx', $staname);
-    $self0_03->call_trigger('item_loaded', $stashkey, \%conf);
-    $self0_03->call_trigger('list_loaded', $stashkey, \%conf);
-    $self0_03->call_trigger('process_request', \$req, \%conf);
-    $self0_03->call_trigger('process_response', \$req, \$res, \%conf);
-    $self0_03->call_trigger('response_decoded', \$req, \$res, \$obj, \%conf);
-    $self0_03->call_trigger('stash_added', $stashkey, \%conf);
-    $self0_03->call_trigger('stash_removed', $stashkey, \%conf);
 
 
 
